part of selene;

/// An event dispatcher, which consumes and emits Discord entity events.
class _Dispatcher {
  /// A map of all event handlers.
  Map<String, Future Function(Map<String, dynamic> data)> _eventHandlers;

  /// The [Client] this dispatcher belongs to.
  Client client;

  // Misc Events

  /// Emitted when the client is ready.
  Stream get onReady => _onReadyController.stream;
  StreamController _onReadyController = new StreamController.broadcast();

  // Guilds

  /// Emitted when the client joins a guild.
  Stream<Guild> get onGuildJoin => _onGuildJoinController.stream;
  StreamController<Guild> _onGuildJoinController =
      new StreamController.broadcast();

  /// Emitted when a guild is updated.
  Stream<Guild> get onGuildUpdate => _onGuildUpdateController.stream;
  StreamController<Guild> _onGuildUpdateController =
      new StreamController.broadcast();

  /// Emitted when the client leaves, or is removed from, a guild.
  Stream<Guild> get onGuildLeave => _onGuildLeaveController.stream;
  StreamController<Guild> _onGuildLeaveController =
      new StreamController.broadcast();

  /// Emitted when a new user joins a guild.
  Stream<GuildMember> get onGuildMemberJoin =>
      _onGuildMemberJoinController.stream;
  StreamController<GuildMember> _onGuildMemberJoinController =
  new StreamController.broadcast();

  /// Emitted when a guild member's information is updated.
  Stream<GuildMember> get onGuildMemberUpdate =>
      _onGuildMemberUpdateController.stream;
  StreamController<GuildMember> _onGuildMemberUpdateController =
  new StreamController.broadcast();

  /// Emitted when a guild member leaves a guild.
  Stream<GuildMember> get onGuildMemberLeave =>
      _onGuildMemberLeaveController.stream;
  StreamController<GuildMember> _onGuildMemberLeaveController =
  new StreamController.broadcast();

  /// Emitted when a role is created in a guild.
  Stream<GuildRole> get onGuildRoleCreate =>
      _onGuildRoleCreateController.stream;
  StreamController<
      GuildRole> _onGuildRoleCreateController = new StreamController
      .broadcast();

  /// Emitted when a role is updated in a guild.
  Stream<GuildRole> get onGuildRoleUpdate =>
      _onGuildRoleUpdateController.stream;
  StreamController<
      GuildRole> _onGuildRoleUpdateController = new StreamController
      .broadcast();

  /// Emitted when a role is deleted from a guild.
  Stream<GuildRole> get onGuildRoleDelete =>
      _onGuildRoleDeleteController.stream;
  StreamController<
      GuildRole> _onGuildRoleDeleteController = new StreamController
      .broadcast();

  // Channels

  /// Emitted when a channel is created that is relevant to the current user.
  Stream<Channel> get onChannelCreate => _onChannelCreateController.stream;
  StreamController<Channel> _onChannelCreateController =
  new StreamController.broadcast();

  /// Emitted when a channel is updated.
  Stream<Channel> get onChannelUpdate => _onChannelUpdateController.stream;
  StreamController<Channel> _onChannelUpdateController =
  new StreamController.broadcast();

  /// Emitted when a channel is deleted.
  Stream<Channel> get onChannelDelete => _onChannelDeleteController.stream;
  StreamController<Channel> _onChannelDeleteController =
  new StreamController.broadcast();

  // Messages

  /// Emitted when a message is received.
  Stream<Message> get onMessageReceive => _onMessageReceiveController.stream;
  StreamController<Message> _onMessageReceiveController =
  new StreamController.broadcast();

  /// Emitted when a message is updated.
  Stream<Message> get onMessageUpdate => _onMessageUpdateController.stream;
  StreamController<Message> _onMessageUpdateController =
  new StreamController.broadcast();

  /// Emitted when a message is deleted.
  Stream get onMessageDelete => _onMessageDeleteController.stream;
  StreamController _onMessageDeleteController =
  new StreamController.broadcast();

  /// Emitted when a bulk array of messages have been deleted.
  Stream<List<Message>> get onMessageBulkDelete =>
      _onMessageBulkDeleteController.stream;
  StreamController<List<Message>> _onMessageBulkDeleteController =
      new StreamController.broadcast();

  /// Creates a new dispatcher, and attaches events.
  _Dispatcher(this.client) {
    _eventHandlers = {
      'READY': (data) async {
        if (data['guilds'] != null) {
          await Future.forEach(data['guilds'], (lazyJsonGuild) async {
            var lazyGuild = new Guild(client);
            lazyGuild._update(lazyJsonGuild);
            client.cache.guilds[lazyGuild.id] = lazyGuild;
          });
        }
        if (data['user'] != null) {
          var newUser = new User(client);
          newUser._update(data['user']);
          client.cache.self = newUser;
          client.cache.users[newUser.id] = newUser; // Cache ourselves
        }
        if (data['private_channels'] != null) {
          await Future.forEach(data['private_channels'],
                  (privateChannelJson) async {
                var privateChannel = new DMChannel(client);
                privateChannel._update(privateChannelJson);
                client.cache.dmChannels[privateChannel.id] = privateChannel;
              });
        }
        _onReadyController.add(null);
      },
      'GUILD_CREATE': (data) async {
        var guildId = int.tryParse(data['id']);
        var cacheResult = client.cache.guilds[guildId];
        if (cacheResult != null) {
          // Lazy loading guild from READY or a guild becomes available again to client
          cacheResult._update(data);
        } else {
          var guild = new Guild(client);
          guild._update(data);
          client.cache.guilds[guildId] = guild;
          _onGuildJoinController.add(guild);
        }
      },
      'GUILD_UPDATE': (data) async {
        var guild = client.cache.guilds[int.tryParse(data['id'])];
        guild._update(data);
        client.cache.guilds[guild.id] = guild;
        _onGuildUpdateController.add(guild);
      },
      'GUILD_DELETE': (data) async {
        var guild = client.cache.guilds[int.tryParse(data['id'])];
        if (data['unavailable'] == null) {
          // Removed or left guild
          client.cache.guilds.remove(int.tryParse(data['id']));
          _onGuildLeaveController.add(guild);
        } else {
          // Guild has gone offline/unavailable
          guild._update(data);
          client.cache.guilds[guild.id] = guild;
        }
      },
      'CHANNEL_CREATE': (data) async {
        var channel = Channel.findImpl(data, client);
        channel._update(data);
        if (channel is GuildChannel) {
          var guildChannel = channel;
          var guild = client.cache.guilds[int.tryParse(data['guild_id'])];
          guildChannel.guild = guild;
          client.cache.guildChannels[guildChannel.id] = guildChannel;
          guild.channels[guildChannel.id] = guildChannel;
        } else if (channel.type == ChannelType.DM) {
          client.cache.dmChannels[channel.id] = channel;
        }
        _onChannelCreateController.add(channel);
      },
      'CHANNEL_UPDATE': (data) async {
        var channel =
        client.cache.resolveChannel(int.tryParse(data['id'] ?? ''));
        channel._update(data);
        _onChannelUpdateController.add(channel);
      },
      'CHANNEL_DELETE': (data) async {
        var parsedId = int.tryParse(data['id'] ?? '');
        var channel = client.cache.resolveChannel(parsedId);
        if (channel == null) return;
        if (channel is GuildChannel) {
          channel.guild.channels.remove(channel.id);
          client.cache.guildChannels.remove(channel.id);
        } else if (channel is DMChannel) {
          client.cache.dmChannels.remove(parsedId);
        }
        _onChannelDeleteController.add(channel);
      },
      'MESSAGE_CREATE': (data) async {
        var message = new Message(client);
        message._update(data);
        message.channel.messages[message.id] = message;

        _onMessageReceiveController.add(message);
      },
      'MESSAGE_UPDATE': (data) async {
        var parsedId = int.tryParse(data['id'] ?? '');
        var parsedChannelId = int.tryParse(data['channel_id'] ?? '');
        var channel =
        client.cache.resolveChannel(parsedChannelId) as TextChannel;
        if (channel == null) return;
        if (channel.messages.containsKey(parsedId)) {
          channel.messages[parsedId]._update(data);
        }
        _onMessageUpdateController.add(channel.messages[parsedId]);
      },
      'MESSAGE_DELETE': (data) async {
        var parsedId = parseInt(data['id']);
        var parsedChannelId = parseInt(data['channel_id']);

        var channel = client.cache.resolveChannel(
            parsedChannelId) as TextChannel;
        if (channel == null) return;
        var message = channel.messages[parsedId];
        channel.messages.remove(parsedId);
        _onMessageDeleteController.add(message);
      },
      'MESSAGE_DELETE_BULK': (data) async {
        var channel =
        client.cache.resolveChannel(int.tryParse(data['channel_id'] ?? ''))
        as TextChannel;
        var list = <Message>[];
        await Future.forEach(data['ids'] ?? [], (jsonMessageId) {
          list.add(channel.messages[parseInt(jsonMessageId)]);
          channel.messages.remove(parseInt(jsonMessageId));
        });
        _onMessageBulkDeleteController.add(list);
      },
      'GUILD_ROLE_CREATE': (data) async {
        var guild = client.cache.guilds[int.tryParse(data['guild_id'])];
        var role = new GuildRole(guild, client);
        role._update(data['role']);
        guild.roles.add(role);
        client.cache.roles[role.id] = role;
        _onGuildRoleCreateController.add(role);
      },
      'GUILD_ROLE_UPDATE': (data) async {
        var guild = client.cache.guilds[int.tryParse(data['guild_id'])];
        var role = guild.roles
            .where((GuildRole r) => r.id == int.tryParse(data['role']['id']))
            .first;
        if (role != null) {
          role._update(data['role']);
          _onGuildRoleUpdateController.add(role);
        }
      },
      'GUILD_ROLE_DELETE': (data) async {
        var guild = client.cache.guilds[int.tryParse(data['guild_id'])];
        var role = guild.roles
            .where((r) => r.id == parseInt(data['role_id']))
            .first;
        guild.roles.remove(role);
        client.cache.roles.remove(role);
        _onGuildRoleDeleteController.add(role);
      },
      'GUILD_MEMBER_ADD': (data) async {
        var guild = client.cache.guilds[parseInt(data['guild_id'])];
        var member = new GuildMember(guild, client);
        member._update(data);
        client.cache.members[member.id] = member;
        guild.members[member.id] = member;
        _onGuildMemberJoinController.add(member);
      },
      'GUILD_MEMBER_UPDATE': (data) async {
        var guild = client.cache.guilds[parseInt(data['guild_id'])];
        var member = guild.members[parseInt(data['user']['id'])];
        if (member != null) {
          member._update(data, true);
          _onGuildMemberUpdateController.add(member);
        }
      },
      'GUILD_MEMBER_REMOVE': (data) async {
        var guild = client.cache.guilds[parseInt(data['guild_id'])];
        var member = guild.members[parseInt(data['user']['id'])];
        guild.members.remove(member.id);
        client.cache.members.remove(member.id);
        _onGuildMemberLeaveController.add(member);
      },
      'GUILD_MEMBERS_CHUNK': (data) async {
        var guild = client.cache.guilds[parseInt(data['guild_id'])];
        await Future.forEach(data['members'], (jsonMember) async {
          var member = new GuildMember(guild, client);
          member._update(jsonMember);
          client.cache.members[member.id] = member;
          guild.members[member.id] = member;
        });
      }
    };
  }

  /// Handles a raw Discord event from Discord.
  Future _handle(String eventName, Map<String, dynamic> data) async {
    var handler = _eventHandlers[eventName];
    if (handler != null) await handler(data);
  }
}
