part of selene;

/// A mixin that enables message CRUD operations.
abstract class TextChannel implements Entity {
  /// Sends a message to this channel.
  ///
  /// Must be connected to the Real-Time Gateway.
  /// Either `content` or `embed` must be provided (or both), otherwise an [ArgumentError] will be raised.
  Future<Message> sendMessage(
      {String content, Map<String, dynamic> embed, bool isTTS = false}) async {
    if (content == null && embed == null) {
      throw new ArgumentError(
          'sendMessage: Content and/or embed must be provided.');
    }

    var requestParams = <String, dynamic>{'tts': isTTS};
    if (content != null) requestParams['content'] = content;
    if (embed != null) requestParams['embed'] = embed;

    var response = await client._rest.createMessage(id, requestParams);
    return (new Message(client))
      .._updateFromResponse(response);
  }

  /// Deletes a message from this channel by ID.
  ///
  /// Client must have the `MANAGE_MESSAGES` permission.
  Future deleteMessageWithId(int messageId) =>
      client._rest.deleteMessage(id, messageId);

  /// Deletes a message from this channel from a message object.
  ///
  /// Client must have the `MANAGE_MESSAGES` permission.
  Future deleteMessage(Message message) => deleteMessageWithId(message.id);

  /// Edits a message by ID.
  ///
  /// Client must be the sender of the message.
  Future<Message> editMessageWithId(int messageId,
      void update(MessageUpdate input)) async {
    var updateObject = new MessageUpdate();
    update(updateObject);
    var response = await client._rest.editMessage(
        id, messageId, updateObject._build());
    return (new Message(client))
      .._updateFromResponse(response);
  }

  /// Edits a message from a message object.
  ///
  /// Client must be the sender of the message.
  Future<Message> editMessage(Message message,
      void update(MessageUpdate input)) =>
      editMessageWithId(message.id, update);

  /// Deletes a group of messages by their IDs.
  ///
  /// Messages cannot be older than two weeks.
  ///
  /// Can only be used in guild channels.
  ///
  /// Requires `MANAGE_MESSAGES `permission.
  Future deleteMessagesById(List<int> ids) async {
    if (ids == null || ids.length < 2 || ids.length > 100) {
      throw new ArgumentError(
          'deleteMessages: ids parameter must be not null and between 2-100 in length');
    }

    await client._rest.deleteMessagesBulk(id, {
      'messages': ids
    });
  }

  /// Deletes a group of messages by objects.
  ///
  /// Messages cannot be older than two weeks.
  ///
  /// Can only be used in guild channels.
  ///
  /// Requires `MANAGE_MESSAGES` permission.
  Future deleteMessages(List<Message> messages) =>
      deleteMessagesById(messages.map((msg) => msg.id));

  /// The ID of the last message sent to this channel.
  int lastMessageId;

  /// The type of this channel.
  ChannelType type;

  /// A message cache.
  Map<int, Message> messages = {};
}
